from openmc_plasma_source import FusionRingSource

import openmc
import pytest
import numpy as np


def test_creation():
    my_source = FusionRingSource(radius=1, z_placement=1)

    # Ensure it is of type openmc.Source
    assert isinstance(my_source, openmc.Source)

    # Ensure it has space, angle, and energy set
    assert isinstance(my_source.space, openmc.stats.CylindricalIndependent)
    assert isinstance(my_source.angle, openmc.stats.Isotropic)
    assert isinstance(my_source.energy, openmc.stats.Muir)


@pytest.mark.parametrize("radius", [1.0, 5.6, 1e5, 7])
def test_radius(radius):
    # should allow any positive float
    my_source = FusionRingSource(radius=radius)
    assert my_source.radius == radius


@pytest.mark.parametrize("radius", [-1.0, "hello world", [1e5]])
def test_bad_radius(radius):
    # should reject any negative float or anything not convertible to float
    with pytest.raises(ValueError):
        my_source = FusionRingSource(radius=radius)


@pytest.mark.parametrize("angles", [(1.0, 2), (0, np.pi), (np.pi, 0)])
def test_angles(angles):
    # Should allow any tuple of length 2 with contents convertible to float
    my_source = FusionRingSource(radius=1.0, angles=angles)
    assert np.array_equal(my_source.angles, angles)


@pytest.mark.parametrize("angles", [(1,), [1, 2], 5, "ab", ("a", "b")])
def test_bad_angles(angles):
    # Should reject iterables of length != 2, anything non tuple, and anything
    # that can't convert to float
    with pytest.raises(ValueError):
        FusionRingSource(radius=1.0, angles=angles)


@pytest.mark.parametrize("temperature", [20000.0, 1e4, 0.1, 25000])
def test_temperature(temperature):
    # Should accept any positive float
    my_source = FusionRingSource(radius=1.0, temperature=temperature)
    assert my_source.temperature == temperature


@pytest.mark.parametrize("temperature", [-20000.0, "hello world", [10000]])
def test_bad_temperature(temperature):
    # Should reject negative floats and anything that isn't convertible to float
    with pytest.raises(ValueError):
        FusionRingSource(radius=1.0, temperature=temperature)


@pytest.mark.parametrize("fuel", ["DT", "DD"])
def test_fuel(fuel):
    # Should accept either 'DD' or 'DT'
    my_source = FusionRingSource(radius=1.0, fuel=fuel)
    assert my_source.fuel_type == fuel


@pytest.mark.parametrize("fuel", ["топливо", 5])
def test_wrong_fuel(fuel):
    # Should reject fuel types besides those listed in fuel_types.py
    with pytest.raises((KeyError, TypeError)):
        FusionRingSource(radius=1.0, fuel=fuel)



from openmc_plasma_source.fuel_types import Fuel, fuel_types
import pytest


@pytest.mark.parametrize("energy,mass", [(2.5e7, 5), (15, 30)])
def test_fuel_with_correct_inputs(energy, mass):
    # Should accept any non-zero positive inputs to either variable
    fuel = Fuel(energy, mass)
    assert fuel.mean_energy == energy
    assert fuel.mass_of_reactants == mass


@pytest.mark.parametrize(
    "energy,mass", [(2.5e7, -5), (-12, 30), (1e7, 0), (0, 4), (-12, -12)]
)
def test_fuel_with_bad_inputs(energy, mass):
    # Should reject any negative numbers and zeros.
    with pytest.raises(ValueError):
        fuel = Fuel(energy, mass)


@pytest.mark.parametrize("fuel_type", ["DT", "DD"])
def test_fuel_types(fuel_type):
    # Should accept 'DD' and 'DT'
    assert isinstance(fuel_types[fuel_type], Fuel)


@pytest.mark.parametrize("fuel_type", ["dt", "dd", "Dt", "dD", "hello world", 5])
def test_incorrect_fuel_types(fuel_type):
    # Should reject everything except 'DT' and 'DD'
    with pytest.raises(KeyError):
        my_fuel = fuel_types[fuel_type]



import openmc
import numpy as np
from typing import Tuple
from param import Parameterized, Number, Range, ListSelector

from .fuel_types import fuel_types


class FusionRingSource(openmc.Source, Parameterized):
    """An openmc.Source object with some presets to make it more convenient
    for fusion simulations using a ring source. All attributes can be changed
    after initialization if required. Default isotropic ring source with a Muir
    energy distribution.

    Args:
        radius (float): the inner radius of the ring source, in metres
        angles (iterable of floats): the start and stop angles of the ring in radians
        z_placement (float): Location of the ring source (m). Defaults to 0.
        temperature (float): the temperature to use in the Muir distribution in eV,
        fuel_type (str): The fusion fuel mix. Either 'DT' or 'DD'.
    """

    radius = Number(None, bounds=(0, None), inclusive_bounds=(False, False))
    angles = Range((0, 2 * np.pi))
    z_placement = Number()
    temperature = Number(bounds=(0, None))
    fuel_type = ListSelector(fuel_types.keys())

    def __init__(
        self,
        radius: float,
        angles: Tuple[float, float] = (0, 2 * np.pi),
        z_placement: float = 0,
        temperature: float = 20000.0,
        fuel: str = "DT",
    ):
        # Set local attributes
        self.radius = radius
        self.angles = angles
        self.z_placement = z_placement
        self.temperature = temperature
        self.fuel_type = fuel
        self.fuel = fuel_types[self.fuel_type]

        # Call init for openmc.Source
        super().__init__()

        # performed after the super init as these are Source attributes
        self.space = openmc.stats.CylindricalIndependent(
            r=openmc.stats.Discrete([self.radius], [1]),
            phi=openmc.stats.Uniform(a=self.angles[0], b=self.angles[1]),
            z=openmc.stats.Discrete([self.z_placement], [1]),
            origin=(0.0, 0.0, 0.0),
        )
        self.angle = openmc.stats.Isotropic()
        self.energy = openmc.stats.Muir(
            e0=self.fuel.mean_energy,
            m_rat=self.fuel.mass_of_reactants,
            kt=self.temperature,
        )



import matplotlib.pyplot as plt
from matplotlib import cm
import numpy as np


def scatter_tokamak_source(source, ax=None, quantity=None, aspect="equal", **kwargs):
    """Create a 2D scatter plot of the tokamak source.
    See https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.scatter.html
    for more arguments.

    Args:
        source (ops.TokamakSource): the plasma source
        ax (maplotlib.Axes, optional): Axes object on which to plot. If not
            provided by the user, the current working Axes is retrieved using
            matplotlib.pyplot.gca().
        aspect (str, optional): Set aspect ratio of the plot. May be set to 'auto',
            'equal', or a number denoting the ratio of the plot height to the plot
            width.
        quantity (str, optional): value by which the lines should be
            coloured. Defaults to None.
        **kwargs: Keyword arguments compatible with matplotlib.pyplot.scatter

    Raises:
        ValueError: If the quantity is unknown
    """

    # Define possible quantities, and link to arrays within tokamak_source
    # If given a non-TokamakSource, this step will fail with an AttributeError
    try:
        quantity_to_attribute = {
            "ion_temperature": source.temperatures,
            "neutron_source_density": source.neutron_source_density,
            "strength": source.strengths,
        }
    except AttributeError as e:
        raise ValueError(
            f"openmc_plasma_source.scatter_tokamak_source: argument 'source' "
            f"must be of type TokamakSource"
        ) from e

    # For a given quantity, determine colours to plot for each point
    # If given an incorrect quantity name, this step will fail with a KeyError
    colours = None
    if quantity is not None:
        try:
            colours = quantity_to_attribute[quantity]
        except KeyError as e:
            raise ValueError(
                f"openmc_plasma_source.scatter_tokamak_source: Unknown 'quantity' "
                f"provided, options are {quantity_to_attribute.keys()}"
            ) from e

    # If not provided with an Axes instance, retrieve the current Axes in focus
    if ax is None:
        ax = plt.gca()

    # Scatter the source R and Z positions, optionally colouring using the chosen
    # quantity.
    ax.scatter(source.RZ[0], source.RZ[1], c=colours, **kwargs)

    # Set the aspect ratio on the axes.
    # Defaults to 'equal', so 1m on the x-axis has the same width as 1m on the y-axis
    ax.set_aspect(aspect)

    return ax


def plot_tokamak_source_3D(
    tokamak_source,
    ax=None,
    quantity=None,
    angles=[0, 0.5 * np.pi],
    colorbar=None,
    **kwargs,
):
    """Creates a 3D plot of the tokamak source.
    See https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.plot.html#matplotlib.pyplot.plot
    for more arguments.

    Args:
        tokamak_source (ops.TokamakSource): the plasma source
        ax (maplotlib.Axes, optional): Axes object on which to plot. If not
            provided by the user, a new Axes is created by calling
            matplotlib.pyplot.axes(projection="3d").
        quantity ("str", optional): value by which the lines should be
            coloured. Defaults to None.
        angles (list, optional): iterable of two floats defining the coverage.
            Defaults to [0, 1/2*np.pi].
        colorbar (str, optional): colorbar used if quantity is not None.
            When None, matplotlib currently defaults to "viridis".

    Raises:
        ValueError: If the quantity is unknown
    """

    # Define possible quantities, and link to arrays within tokamak_source
    # If given a non-TokamakSource, this step will fail with an AttributeError
    try:
        quantity_to_attribute = {
            "ion_temperature": tokamak_source.temperatures,
            "neutron_source_density": tokamak_source.neutron_source_density,
            "strength": tokamak_source.strengths,
        }
    except AttributeError as e:
        raise ValueError(
            f"openmc_plasma_source.plot_tokamak_source_3D: argument 'source' "
            f"must be of type TokamakSource"
        ) from e

    # For a given quantity, get the corresponding array from tokamak_source
    # If given an incorrect quantity name, this step will fail with a KeyError
    if quantity is not None:
        try:
            quantity_values = quantity_to_attribute[quantity]
        except KeyError as e:
            raise ValueError(
                f"openmc_plasma_source.plot_tokamak_source_3D: Unknown 'quantity' "
                f"provided, options are {quantity_to_attribute.keys()}"
            ) from e
    else:
        quantity_values = np.ones(tokamak_source.sample_size)

    # Get the colours used to plot each curve
    # If 'quantity' == None, all have the same colour, selected from the middle
    # of the colormap.
    cmap = cm.get_cmap(colorbar)
    if quantity is not None:
        colors = cmap(quantity_values / np.max(quantity_values))
    else:
        colors = cmap(np.full(tokamak_source.sample_size, 0.5))

    # If not provided with an Axes object, create a new one
    if ax is None:
        ax = plt.axes(projection="3d")

    # Get curves to plot
    n_theta = 100
    theta = np.linspace(*angles, n_theta)
    xs = np.outer(tokamak_source.RZ[0], np.sin(theta))
    ys = np.outer(tokamak_source.RZ[0], np.cos(theta))
    zs = tokamak_source.RZ[1]

    # Plot each curve in turn
    for x, y, z, color in zip(xs, ys, zs, colors):
        ax.plot(x, y, z, color=color, **kwargs)

    # Set plot bounds
    major_radius = tokamak_source.major_radius
    ax.set_xlim(-major_radius, major_radius)
    ax.set_ylim(-major_radius, major_radius)
    ax.set_zlim(-major_radius, major_radius)

    return ax



"""fuel_types.py

Defines dictionary for determining mean energy and mass of reactants
for a given fusion fuel type.
"""

from param import Parameterized, Number


class Fuel(Parameterized):
    # mean energy, eV
    mean_energy = Number(None, bounds=(0, None), inclusive_bounds=(False, False))

    # mass of the reactants, AMU
    mass_of_reactants = Number(None, bounds=(0, None), inclusive_bounds=(False, False))

    def __init__(self, mean_energy, mass_of_reactants):
        self.mean_energy = mean_energy
        self.mass_of_reactants = mass_of_reactants


fuel_types = {
    "DD": Fuel(mean_energy=2450000.0, mass_of_reactants=4),
    "DT": Fuel(mean_energy=14080000.0, mass_of_reactants=5),
}



import matplotlib.pyplot as plt
import numpy as np
from openmc_plasma_source import (
    TokamakSource,
    plotting as ops_plt,
)
import pytest


@pytest.fixture
def tokamak_source():
    return TokamakSource(
        elongation=1.557,
        ion_density_centre=1.09e20,
        ion_density_peaking_factor=1,
        ion_density_pedestal=1.09e20,
        ion_density_separatrix=3e19,
        ion_temperature_centre=45.9,
        ion_temperature_peaking_factor=8.06,
        ion_temperature_pedestal=6.09,
        ion_temperature_separatrix=0.1,
        major_radius=9.06,
        minor_radius=2.92258,
        pedestal_radius=0.8 * 2.92258,
        mode="H",
        shafranov_factor=0.44789,
        triangularity=0.270,
        ion_temperature_beta=6,
        sample_size=500,
    )


def test_scatter_tokamak_source_defaults(tokamak_source):
    """Ensure plotting is successful without providing additional args"""
    plt.figure()
    assert not plt.gca().collections  # Check current ax is empty
    ops_plt.scatter_tokamak_source(tokamak_source)
    assert plt.gca().collections  # Check current ax is not empty
    # Save for viewing, clean up
    plt.xlabel("R")
    plt.ylabel("Z", rotation=0)
    plt.savefig("tests/test_scatter_tokamak_source_defaults.png")
    plt.close()


def test_scatter_tokamak_source_with_ax(tokamak_source):
    """Ensure plotting is successful for user-provided ax"""
    fig = plt.figure()
    ax = fig.gca()
    assert not ax.collections  # Check ax is empty
    ops_plt.scatter_tokamak_source(tokamak_source, ax=ax)
    assert ax.collections  # Check ax is not empty
    # Save for viewing, clean up
    ax.set_xlabel("R")
    ax.set_ylabel("Z", rotation=0)
    fig.savefig("tests/test_scatter_tokamak_source_with_ax.png")
    plt.close(fig)


def test_scatter_tokamak_source_with_subplots(tokamak_source):
    """Ensure plotting is successful for multiple user-provided ax"""
    fig, (ax1, ax2) = plt.subplots(1, 2)
    # Plot on the first axes
    assert not ax1.collections  # Check ax is empty
    ops_plt.scatter_tokamak_source(tokamak_source, ax=ax1)
    assert ax1.collections  # Check ax is not empty
    # Generate new data
    tokamak_source.sample_sources()
    tokamak_source.sources = tokamak_source.make_openmc_sources()
    # Plot on the other axes
    assert not ax2.collections  # Check ax is empty
    ops_plt.scatter_tokamak_source(tokamak_source, ax=ax2)
    assert ax2.collections  # Check ax is not empty
    # Save for viewing, clean up
    ax1.set_xlabel("R")
    ax1.set_ylabel("Z", rotation=0)
    ax2.set_xlabel("R")
    fig.savefig("tests/test_scatter_tokamak_source_subplots.png")
    plt.close(fig)


@pytest.mark.parametrize(
    "quantity", ["ion_temperature", "neutron_source_density", "strength"]
)
def test_scatter_tokamak_source_quantities(tokamak_source, quantity):
    """Plot with colours set by 'quantity'"""
    fig = plt.figure()
    ax = fig.gca()
    assert not ax.collections  # Check ax is empty
    ops_plt.scatter_tokamak_source(tokamak_source, ax=ax, quantity=quantity)
    assert ax.collections  # Check ax is not empty
    # Save for viewing, clean up
    ax.set_xlabel("R")
    ax.set_ylabel("Z", rotation=0)
    fig.savefig(f"tests/test_scatter_tokamak_source_quantities_{quantity}.png")
    plt.close(fig)


@pytest.mark.parametrize("aspect", ["equal", "auto", 2])
def test_scatter_tokamak_source_aspect(tokamak_source, aspect):
    """Plot with various aspect ratios"""
    fig = plt.figure()
    ax = fig.gca()
    assert not ax.collections  # Check ax is empty
    ops_plt.scatter_tokamak_source(tokamak_source, ax=ax, aspect=aspect)
    assert ax.collections  # Check ax is not empty
    # Save for viewing, clean up
    ax.set_xlabel("R")
    ax.set_ylabel("Z", rotation=0)
    fig.savefig(f"tests/test_scatter_tokamak_source_aspect_{aspect}.png")
    plt.close(fig)


@pytest.mark.parametrize("kwargs", [{"alpha": 0.2}, {"marker": "x"}])
def test_scatter_tokamak_source_kwargs(tokamak_source, kwargs):
    """Plot with a kwarg compatible with 'scatter'"""
    fig = plt.figure()
    ax = fig.gca()
    assert not ax.collections  # Check ax is empty
    ops_plt.scatter_tokamak_source(tokamak_source, ax=ax, **kwargs)
    assert ax.collections  # Check ax is not empty
    # Save for viewing, clean up
    ax.set_xlabel("R")
    ax.set_ylabel("Z", rotation=0)
    fig.savefig(
        f"tests/test_scatter_tokamak_source_kwargs_{list(kwargs.keys())[0]}.png"
    )
    plt.close(fig)


def test_scatter_tokamak_not_source():
    """Ensure failure when given non-TokamakSource to plot"""
    with pytest.raises(ValueError) as excinfo:
        fig = plt.figure()
        ax = fig.gca()
        ops_plt.scatter_tokamak_source("hello world", ax=ax)
    plt.close()
    assert "TokamakSource" in str(excinfo.value)


@pytest.mark.parametrize("quantity", ["coucou", "ion_density", 17])
def test_scatter_tokamak_wrong_quantity(tokamak_source, quantity):
    """Ensure failure when incorrect quantity specified"""
    with pytest.raises(ValueError) as excinfo:
        fig = plt.figure()
        ax = fig.gca()
        ops_plt.scatter_tokamak_source(tokamak_source, ax=ax, quantity=quantity)
    plt.close()
    assert "quantity" in str(excinfo.value)


def test_plot_tokamak_source_3D_default(tokamak_source):
    """Ensure plots correctly with default inputs"""
    plt.figure()
    ops_plt.plot_tokamak_source_3D(tokamak_source)
    assert plt.gca().lines  # Check current ax is not empty
    # Save for viewing, clean up
    plt.savefig("tests/test_plot_tokamak_source_3D_defaults.png")
    plt.close()


def test_plot_tokamak_source_3D_with_ax(tokamak_source):
    """Ensure plots correctly given ax instance"""
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1, projection="3d")
    assert not ax.lines  # Check ax is empty
    ops_plt.plot_tokamak_source_3D(tokamak_source, ax=ax)
    assert ax.lines  # Check ax is not empty
    # Save for viewing, clean up
    fig.savefig("tests/test_plot_tokamak_source_3D_with_ax.png")
    plt.close(fig)


@pytest.mark.parametrize(
    "quantity", ["ion_temperature", "neutron_source_density", "strength"]
)
def test_plot_tokamak_source_3D_quantities(tokamak_source, quantity):
    """Ensure plots correctly for each quantity"""
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1, projection="3d")
    assert not ax.lines  # Check ax is empty
    ops_plt.plot_tokamak_source_3D(tokamak_source, ax=ax, quantity=quantity)
    assert ax.lines  # Check ax is not empty
    # Save for viewing, clean up
    fig.savefig(f"tests/test_plot_tokamak_source_3D_quantities_{quantity}.png")
    plt.close(fig)


@pytest.mark.parametrize("colorbar", ["plasma", "rainbow"])
def test_plot_tokamak_source_3D_colorbars(tokamak_source, colorbar):
    """Ensure plots correctly given colorbar choice"""
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1, projection="3d")
    assert not ax.lines  # Check ax is empty
    ops_plt.plot_tokamak_source_3D(
        tokamak_source, ax=ax, quantity="ion_temperature", colorbar=colorbar
    )
    assert ax.lines  # Check ax is not empty
    # Save for viewing, clean up
    fig.savefig(f"tests/test_plot_tokamak_source_3D_colorbar_{colorbar}.png")
    plt.close(fig)


@pytest.mark.parametrize(
    "angles,name",
    [
        [(0, np.pi / 4), "eighth"],
        [(0, np.pi), "half"],
        [(np.pi, 2 * np.pi), "half_offset"],
        [(0, 2 * np.pi), "full"],
    ],
)
def test_plot_tokamak_source_3D_angles(tokamak_source, angles, name):
    """Ensure plots correctly given angles range"""
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1, projection="3d")
    assert not ax.lines  # Check ax is empty
    ops_plt.plot_tokamak_source_3D(
        tokamak_source, ax=ax, quantity="ion_temperature", angles=angles
    )
    assert ax.lines  # Check ax is not empty
    # Save for viewing, clean up
    fig.savefig(f"tests/test_plot_tokamak_source_3D_angles_{name}.png")
    plt.close(fig)


@pytest.mark.parametrize("kwargs", [{"alpha": 0.2}, {"linestyle": "dotted"}])
def test_plot_tokamak_source_3D_kwargs(tokamak_source, kwargs):
    """Ensure plots correctly given additonal keyword arguments to pass on to plot"""
    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1, projection="3d")
    assert not ax.lines  # Check ax is empty
    ops_plt.plot_tokamak_source_3D(
        tokamak_source, ax=ax, quantity="ion_temperature", **kwargs
    )
    assert ax.lines  # Check ax is not empty
    # Save for viewing, clean up
    fig.savefig(
        f"tests/test_plot_tokamak_source_3D_kwargs_{list(kwargs.keys())[0]}.png"
    )
    plt.close(fig)


def test_plot_tokamak_source_3D_not_source():
    """Ensure failure when given non-TokamakSource to plot"""
    with pytest.raises(ValueError) as excinfo:
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1, projection="3d")
        ops_plt.plot_tokamak_source_3D("hello world", ax=ax)
    plt.close()
    assert "TokamakSource" in str(excinfo.value)


@pytest.mark.parametrize("quantity", ["coucou", "ion_density", 17])
def test_plot_tokamak_source_3D_wrong_quantity(tokamak_source, quantity):
    """Ensure failure when incorrect quantity specified"""
    with pytest.raises(ValueError) as excinfo:
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1, projection="3d")
        ops_plt.plot_tokamak_source_3D(tokamak_source, ax=ax, quantity=quantity)
    plt.close()
    assert "quantity" in str(excinfo.value)



import openmc
from typing import Tuple
from param import Parameterized, Number, NumericTuple, ListSelector

from .fuel_types import fuel_types


class FusionPointSource(openmc.Source, Parameterized):
    """An openmc.Source object with some presets to make it more convenient
    for fusion simulations using a point source. All attributes can be changed
    after initialization if required. Default isotropic point source at the
    origin with a Muir energy distribution.

    Args:
        coordinate (tuple[float,float,float]): Location of the point source.
            Each component is measured in metres.
        temperature (float): Temperature of the source (eV).
        fuel_type (str): The fusion fuel mix. Either 'DT' or 'DD'.
    """

    coordinate = NumericTuple(None, length=3)
    temperature = Number(None, bounds=(0, None))  # temperature in eV
    fuel_type = ListSelector(fuel_types.keys())

    def __init__(
        self,
        coordinate: Tuple[float, float, float] = (0, 0, 0),
        temperature: float = 20000.0,
        fuel: str = "DT",
    ):
        # Set local attributes
        self.coordinate = coordinate
        self.temperature = temperature
        self.fuel_type = fuel
        self.fuel = fuel_types[self.fuel_type]

        # Call init for openmc.Source
        super().__init__()

        # performed after the super init as these are Source attributes
        self.space = openmc.stats.Point(self.coordinate)
        self.angle = openmc.stats.Isotropic()
        self.energy = openmc.stats.Muir(
            e0=self.fuel.mean_energy,
            m_rat=self.fuel.mass_of_reactants,
            kt=self.temperature,
        )



from openmc_plasma_source import TokamakSource
from openmc import Source
import numpy as np

import pytest
from hypothesis import given, settings, assume, strategies as st


@pytest.fixture
def tokamak_args_dict():
    """Returns a dict of realistic inputs for TokamakSource"""
    args_dict = {
        "elongation": 1.557,
        "triangularity": 0.270,
        "major_radius": 9.06,
        "minor_radius": 2.92258,
        "pedestal_radius": 0.8 * 2.92258,
        "shafranov_factor": 0.44789,
        "ion_density_centre": 1.09e20,
        "ion_density_peaking_factor": 1,
        "ion_density_pedestal": 1.09e20,
        "ion_density_separatrix": 3e19,
        "ion_temperature_centre": 45.9,
        "ion_temperature_peaking_factor": 8.06,
        "ion_temperature_pedestal": 6.09,
        "ion_temperature_separatrix": 0.1,
        "mode": "H",
        "ion_temperature_beta": 6,
    }
    return args_dict


@pytest.fixture
def tokamak_source_example(tokamak_args_dict):
    """Returns a TokamakSource with realistic inputs"""
    return TokamakSource(**tokamak_args_dict)


def test_creation(tokamak_source_example):
    """Tests that the sources generated by TokamakSource are of
    type openmc.Source"""
    for source in tokamak_source_example.sources:
        assert isinstance(source, Source)


@pytest.mark.parametrize(
    "minor_radius,major_radius", [(3.0, 10.0), (3.0, 100), (3.0, 3.00001)]
)
def test_major_radius(tokamak_args_dict, minor_radius, major_radius):
    """Checks that TokamakSource creation accepts valid major radius"""
    tokamak_args_dict["minor_radius"] = minor_radius
    tokamak_args_dict["major_radius"] = major_radius
    tokamak_source = TokamakSource(**tokamak_args_dict)
    assert tokamak_source.major_radius == major_radius


@pytest.mark.parametrize(
    "minor_radius,major_radius", [(3, 3), (3, 1), (3, -5), (3, "hello world")]
)
def test_bad_major_radius(tokamak_args_dict, minor_radius, major_radius):
    """Checks that TokamakSource creation rejects invalid major radius"""
    tokamak_args_dict["minor_radius"] = minor_radius
    tokamak_args_dict["major_radius"] = major_radius
    with pytest.raises(ValueError):
        tokamak_source = TokamakSource(**tokamak_args_dict)


@pytest.mark.parametrize(
    "major_radius,minor_radius", [(10.0, 3.0), (10.0, 9.9), (10.0, 0.1)]
)
def test_minor_radius(tokamak_args_dict, major_radius, minor_radius):
    """Checks that TokamakSource creation accepts valid minor radius"""
    tokamak_args_dict["major_radius"] = major_radius
    tokamak_args_dict["minor_radius"] = minor_radius
    # Set shafranov factor to 0 and pedestal factor to 0.8*minor_radius for safety
    tokamak_args_dict["pedestal_radius"] = 0.8 * minor_radius
    tokamak_args_dict["shafranov_factor"] = 0.0
    tokamak_source = TokamakSource(**tokamak_args_dict)
    assert tokamak_source.minor_radius == minor_radius


@pytest.mark.parametrize(
    "major_radius,minor_radius",
    [(10.0, 10.0), (10.0, 20.0), (10.0, 0), (10.0, -6), (10.0, "hello world")],
)
def test_bad_minor_radius(tokamak_args_dict, major_radius, minor_radius):
    """Checks that TokamakSource creation rejects invalid minor radius"""
    tokamak_args_dict["major_radius"] = major_radius
    tokamak_args_dict["minor_radius"] = minor_radius
    with pytest.raises(ValueError):
        tokamak_source = TokamakSource(**tokamak_args_dict)


@pytest.mark.parametrize("elongation", [1.0, 1.667, 0.5, 20, 0.001])
def test_elongation(tokamak_args_dict, elongation):
    """Checks that TokamakSource creation accepts valid elongation"""
    tokamak_args_dict["elongation"] = elongation
    tokamak_source = TokamakSource(**tokamak_args_dict)
    assert tokamak_source.elongation == elongation


@pytest.mark.parametrize("elongation", [0, -5, "hello world"])
def test_bad_elongation(tokamak_args_dict, elongation):
    """Checks that TokamakSource creation rejects invalid elongation"""
    tokamak_args_dict["elongation"] = elongation
    with pytest.raises(ValueError):
        tokamak_source = TokamakSource(**tokamak_args_dict)


@pytest.mark.parametrize("triangularity", [0.0, 0.5, 0.9, 1.0, -0.5, -0.9, -1.0])
def test_triangularity(tokamak_args_dict, triangularity):
    """Checks that TokamakSource creation accepts valid triangularity"""
    tokamak_args_dict["triangularity"] = triangularity
    tokamak_source = TokamakSource(**tokamak_args_dict)
    assert tokamak_source.triangularity == triangularity


@pytest.mark.parametrize("triangularity", [1.1, -1.1, 10, -10, "hello world"])
def test_bad_triangularity(tokamak_args_dict, triangularity):
    """Checks that TokamakSource creation rejects invalid triangularity"""
    tokamak_args_dict["triangularity"] = triangularity
    with pytest.raises(ValueError):
        tokamak_source = TokamakSource(**tokamak_args_dict)


@pytest.mark.parametrize(
    "major_radius,minor_radius,shaf",
    [
        (10.0, 3.0, 0),
        (10.0, 3.0, 1.0),
        (10.0, 3.0, -1.0),
        (10.0, 3.0, 1.49),
        (10.0, 3.0, -1.49),
        (10.0, 5.0, 2.49),
        (10.0, 5.0, -2.49),
    ],
)
def test_shafranov_factor(tokamak_args_dict, major_radius, minor_radius, shaf):
    """Checks that TokamakSource creation accepts valid Shafranov factor"""
    tokamak_args_dict["major_radius"] = major_radius
    tokamak_args_dict["minor_radius"] = minor_radius
    tokamak_args_dict["pedestal_radius"] = 0.8 * minor_radius
    tokamak_args_dict["shafranov_factor"] = shaf
    tokamak_source = TokamakSource(**tokamak_args_dict)
    assert tokamak_source.shafranov_factor == shaf


@pytest.mark.parametrize(
    "major_radius,minor_radius,shaf",
    [
        (10.0, 3.0, 3.0),
        (10.0, 3.0, -3.0),
        (10.0, 3.0, 1.5),
        (10.0, 3.0, -1.5),
        (10.0, 5.0, 2.5),
        (10.0, 5.0, -2.5),
    ],
)
def test_bad_shafranov_factor(tokamak_args_dict, major_radius, minor_radius, shaf):
    """Checks that TokamakSource creation rejects invalid Shafranov factor"""
    tokamak_args_dict["major_radius"] = major_radius
    tokamak_args_dict["minor_radius"] = minor_radius
    tokamak_args_dict["pedestal_radius"] = 0.8 * minor_radius
    tokamak_args_dict["shafranov_factor"] = shaf
    with pytest.raises((ValueError, TypeError)):
        tokamak_source = TokamakSource(**tokamak_args_dict)


@pytest.mark.parametrize(
    "angles", [(0, 1), (-np.pi, np.pi), (3 * np.pi / 4, -np.pi / 4)]
)
def test_angles(tokamak_args_dict, angles):
    """Checks that custom angles can be set"""
    # Note: should accept negative angles and angles in reverse order
    tokamak_args_dict["angles"] = angles
    tokamak_source = TokamakSource(**tokamak_args_dict)
    assert np.array_equal(tokamak_source.angles, angles)
    for source in tokamak_source.sources:
        assert np.array_equal((source.space.phi.a, source.space.phi.b), angles)


@pytest.mark.parametrize("angles", [(0, 1, 2), -5, ("hello", "world")])
def test_bad_angles(tokamak_args_dict, angles):
    """Checks that invalid angles are rejected"""
    # It should fail when given something that isn't a 2-tuple or similar
    # Contents should convert to float
    tokamak_args_dict["angles"] = angles
    with pytest.raises(ValueError) as excinfo:
        tokamak_source = TokamakSource(**tokamak_args_dict)


def test_ion_density(tokamak_source_example):
    # test with values of r that are within acceptable ranges.
    r = np.linspace(0.0, tokamak_source_example.minor_radius, 100)
    density = tokamak_source_example.ion_density(r)
    assert isinstance(r, np.ndarray)
    assert len(density) == len(r)
    assert np.all(np.isfinite(density))


def test_bad_ion_density(tokamak_source_example):
    # It should fail if given a negative r
    with pytest.raises(ValueError) as excinfo:
        density = tokamak_source_example.ion_density([0, 5, -6])
    assert "must not be negative" in str(excinfo.value)


def test_ion_temperature(tokamak_source_example):
    # test with values of r that are within acceptable ranges.
    r = np.linspace(0.0, tokamak_source_example.minor_radius, 100)
    temperature = tokamak_source_example.ion_temperature(r)
    assert isinstance(temperature, np.ndarray)
    assert len(temperature) == len(r)
    assert np.all(np.isfinite(temperature))


def test_bad_ion_temperature(tokamak_source_example):
    # It should fail if given a negative r
    with pytest.raises(ValueError) as excinfo:
        temperature = tokamak_source_example.ion_temperature([0, 5, -6])
    assert "must not be negative" in str(excinfo.value)


def test_convert_a_alpha_to_R_Z(tokamak_source_example):
    # Similar to  test_source_locations_are_within_correct_range
    # Rather than going in detail, simply tests validity of inputs and outputs
    # Test with suitable values for a and alpha
    a = np.linspace(0.0, tokamak_source_example.minor_radius, 100)
    alpha = np.linspace(0.0, 2 * np.pi, 100)
    R, Z = tokamak_source_example.convert_a_alpha_to_R_Z(a, alpha)
    assert isinstance(R, np.ndarray)
    assert isinstance(Z, np.ndarray)
    assert len(R) == len(a)
    assert len(Z) == len(a)
    assert np.all(np.isfinite(R))
    assert np.all(np.isfinite(Z))


def test_bad_convert_a_alpha_to_R_Z(tokamak_source_example):
    # Repeat test_convert_a_alpha_to_R_Z, but show that negative a breaks it
    a = np.linspace(0.0, tokamak_source_example.minor_radius, 100)
    alpha = np.linspace(0.0, 2 * np.pi, 100)
    with pytest.raises(ValueError) as excinfo:
        R, Z = tokamak_source_example.convert_a_alpha_to_R_Z(-a, alpha)
    assert "must not be negative" in str(excinfo.value)


@st.composite
def tokamak_source_strategy(draw):
    """Defines a hypothesis strategy that automatically generates a TokamakSource.
    Geometry attributes are varied, while plasma attributes are fixed.
    """
    # Used to avoid generation of inappropriate float values
    finites = {
        "allow_nan": False,
        "allow_infinity": False,
        "allow_subnormal": False,
    }

    # Specify the base strategies for each geometry input
    major_radius = draw(st.floats(min_value=1e-5, max_value=100.0, **finites))

    minor_radius = draw(
        st.floats(
            min_value=1e-5 * major_radius,
            max_value=np.nextafter(major_radius, major_radius - 1),
            **finites
        )
    )

    pedestal_radius = draw(
        st.floats(
            min_value=0.8 * minor_radius,
            max_value=np.nextafter(minor_radius, minor_radius - 1),
            **finites
        )
    )

    elongation = draw(st.floats(min_value=1e-5, max_value=10.0, **finites))

    triangularity = draw(
        st.floats(
            min_value=np.nextafter(-1.0, +1), max_value=np.nextafter(1.0, -1), **finites
        )
    )

    shafranov_factor = draw(
        st.floats(
            min_value=np.nextafter(-0.5 * minor_radius, +1),
            max_value=np.nextafter(0.5 * minor_radius, -1),
            **finites
        )
    )

    return TokamakSource(
        elongation=elongation,
        triangularity=triangularity,
        major_radius=major_radius,
        minor_radius=minor_radius,
        pedestal_radius=pedestal_radius,
        shafranov_factor=shafranov_factor,
        ion_density_centre=1.09e20,
        ion_density_peaking_factor=1,
        ion_density_pedestal=1.01e20,
        ion_density_separatrix=3e19,
        ion_temperature_centre=45.9,
        ion_temperature_peaking_factor=8.06,
        ion_temperature_pedestal=6.09,
        ion_temperature_separatrix=0.1,
        mode="H",
        ion_temperature_beta=6,
    )


@given(tokamak_source=tokamak_source_strategy())
@settings(max_examples=50)
def test_strengths_are_normalised(tokamak_source):
    """Tests that the sum of the strengths attribute is equal to"""
    assert pytest.approx(sum(tokamak_source.strengths)) == 1


@given(tokamak_source=tokamak_source_strategy())
@settings(max_examples=50)
def test_source_locations_are_within_correct_range(tokamak_source):
    """Tests that each source has RZ locations within the expected range.

    As the function converting (a,alpha) coordinates to (R,Z) is not bijective,
    we cannot convert back to validate each individual point. However, we can
    determine whether the generated points are contained within the shell of
    the last closed magnetic surface.  See "Tokamak D-T neutron source models
    for different plasma physics confinement modes", C. Fausser et al., Fusion
    Engineering and Design, 2012 for more info.
    """
    R_0 = tokamak_source.major_radius
    A = tokamak_source.minor_radius
    El = tokamak_source.elongation
    delta = tokamak_source.triangularity

    def get_R_on_LCMS(alpha):
        """Gets R on the last closed magnetic surface for a given alpha"""
        return R_0 + A * np.cos(alpha + delta * np.sin(alpha))

    approx_lt = lambda x, y: x < y or np.isclose(x, y)
    approx_gt = lambda x, y: x > y or np.isclose(x, y)

    for source in tokamak_source.sources:
        R, Z = source.space.r.x[0], source.space.z.x[0]
        # First test that the point is contained with a simple box with
        # lower left (r_min,-z_max) and upper right (r_max,z_max)
        assert approx_gt(R, R_0 - A)
        assert approx_lt(R, R_0 + A)
        assert approx_lt(abs(Z), A * El)
        # For a given Z, we can determine the two values of alpha where
        # where a = minor_radius, and from there determine the upper and
        # lower bounds for R.
        alpha_1 = np.arcsin(abs(Z) / (El * A))
        alpha_2 = np.pi - alpha_1
        R_max, R_min = get_R_on_LCMS(alpha_1), get_R_on_LCMS(alpha_2)
        assert approx_lt(R_max, R_0 + A)
        assert approx_gt(R_min, R_0 - A)
        assert approx_lt(R, R_max)
        assert approx_gt(R, R_min)



from openmc_plasma_source import FusionPointSource

import openmc
import pytest
import numpy as np


def test_creation():
    my_source = FusionPointSource()

    # Ensure it is of type openmc.Source
    assert isinstance(my_source, openmc.Source)

    # Ensure it has space, angle, and energy set
    assert isinstance(my_source.space, openmc.stats.Point)
    assert isinstance(my_source.angle, openmc.stats.Isotropic)
    assert isinstance(my_source.energy, openmc.stats.Muir)


@pytest.mark.parametrize(
    "coordinate", [(1.0, 2.0, 3.0), (4, 5, 6), tuple(np.linspace(1.0, 3.0, 3))]
)
def test_coordinate(coordinate):
    # Should allow any tuple of length 3 containing numbers
    my_source = FusionPointSource(coordinate=coordinate)
    assert np.array_equal(my_source.coordinate, coordinate)


@pytest.mark.parametrize("coordinate", [(1, 2), [3, 4, 5], 5, "abc", ("a", "b", "c")])
def test_bad_coordinate(coordinate):
    # Should reject iterables of length != 3, anything non-tuple, and anything
    # that can't convert to float
    with pytest.raises(ValueError):
        FusionPointSource(coordinate=coordinate)


@pytest.mark.parametrize("temperature", [20000.0, 1e4, 0.1, 25000])
def test_temperature(temperature):
    # Should accept any positive float
    my_source = FusionPointSource(temperature=temperature)
    assert my_source.temperature == temperature


@pytest.mark.parametrize("temperature", [-20000.0, "hello world", [10000]])
def test_bad_temperature(temperature):
    # Should reject negative floats and anything that isn't convertible to float
    with pytest.raises(ValueError):
        FusionPointSource(temperature=temperature)


@pytest.mark.parametrize("fuel", ["DT", "DD"])
def test_fuel(fuel):
    # Should accept either 'DD' or 'DT'
    my_source = FusionPointSource(fuel=fuel)
    assert my_source.fuel_type == fuel


@pytest.mark.parametrize("fuel", ["топливо", 5])
def test_wrong_fuel(fuel):
    # Should reject fuel types besides those listed in fuel_types.py
    with pytest.raises((KeyError, TypeError)):
        FusionPointSource(fuel=fuel)
